/*
 * Copyright (C) 2007 by egnite Software GmbH. All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. Neither the name of the copyright holders nor the names of
 *    contributors may be used to endorse or promote products derived
 *    from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY EGNITE SOFTWARE GMBH AND CONTRIBUTORS
 * ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 * FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL EGNITE
 * SOFTWARE GMBH OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS
 * OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED
 * AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF
 * THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * For additional information see http://www.ethernut.de/
 *
 */

/*!
 * \file buttons.c
 * \brief Simple button interface.
 *
 * Handles 3 toggle switches connected to 3 dedicated GPIO lines. It is
 * expected, that the port pin is driven low when the related button is
 * pressed.
 *
 * Auto repeat is supported.
 *
 * \verbatim
 *
 * $Log$
 *
 * \endverbatim
 */

#include <cfg/os.h>
#include <cfg/clock.h>
#include <arch/arm.h>
#include <dev/irqreg.h>

#include <sys/event.h>
#include <sys/timer.h>

#include "buttons.h"

#if defined(AT91SAM7X_EK)
#define BTN_TC_ID       TC1_ID
#define BTN_PIO_ID      PIOB_ID
#define BTN_DOWN_BIT    27
//#define BTN_MENU_BIT    28 //Doesn't work. Reason currently unknown.
#define BTN_UP_BIT      29
#elif defined(AT91SAM9260_EK)
#define BTN_TC_ID       TC1_ID
#define BTN_PIO_ID      PIOB_ID
#define BTN_DOWN_BIT    8
#define BTN_MENU_BIT    9
#define BTN_UP_BIT      10
#endif

#ifndef BTN_SCAN_FREQ
/*! \brief Number of button scans per second. */
#define BTN_SCAN_FREQ       128UL
#endif

#ifndef BTN_REPEAT_FIRST
/*! \brief Number of scans until the first repeat is triggered. */
#define BTN_REPEAT_FIRST    100
#endif

#ifndef BTN_REPEAT_NEXT
/*! \brief Number of scans until all following repeats are triggered. */
#define BTN_REPEAT_NEXT     25
#endif

#if BTN_PIO_ID == PIOA_ID
#define BTN_PIO_PE_REG  PIOA_PER
#define BTN_PIO_OE_REG  PIOA_OER
#define BTN_PIO_OD_REG  PIOA_ODR
#define BTN_PIO_SOD_REG PIOA_SODR
#define BTN_PIO_COD_REG PIOA_CODR
#define BTN_PIO_PDS_REG PIOA_PDSR
#define BTN_PIO_PUE_REG PIOA_PUER
#define BTN_PIO_PUD_REG PIOA_PUDR
#elif BTN_PIO_ID == PIOB_ID
#define BTN_PIO_PE_REG  PIOB_PER
#define BTN_PIO_OE_REG  PIOB_OER
#define BTN_PIO_OD_REG  PIOB_ODR
#define BTN_PIO_SOD_REG PIOB_SODR
#define BTN_PIO_COD_REG PIOB_CODR
#define BTN_PIO_PDS_REG PIOB_PDSR
#define BTN_PIO_PUE_REG PIOB_PUER
#define BTN_PIO_PUD_REG PIOB_PUDR
#elif BTN_PIO_ID == PIOC_ID
#define BTN_PIO_PE_REG  PIOC_PER
#define BTN_PIO_OE_REG  PIOC_OER
#define BTN_PIO_OD_REG  PIOC_ODR
#define BTN_PIO_SOD_REG PIOC_SODR
#define BTN_PIO_COD_REG PIOC_CODR
#define BTN_PIO_PDS_REG PIOC_PDSR
#define BTN_PIO_PUE_REG PIOC_PUER
#define BTN_PIO_PUD_REG PIOC_PUDR
#else
#define BTN_PIO_PE_REG  PIO_PER
#define BTN_PIO_OE_REG  PIO_OER
#define BTN_PIO_OD_REG  PIO_ODR
#define BTN_PIO_SOD_REG PIO_SODR
#define BTN_PIO_COD_REG PIO_CODR
#define BTN_PIO_PDS_REG PIO_PDSR
#define BTN_PIO_PUE_REG PIO_PUER
#define BTN_PIO_PUD_REG PIO_PUDR
#endif

#ifdef BTN_MENU_BIT
#define BTN_MENU        _BV(BTN_MENU_BIT)
#else
#define BTN_MENU        0
#endif
#ifdef BTN_UP_BIT
#define BTN_UP          _BV(BTN_UP_BIT)
#else
#define BTN_UP          0
#endif
#ifdef BTN_DOWN_BIT
#define BTN_DOWN        _BV(BTN_DOWN_BIT)
#else
#define BTN_DOWN        0
#endif

#if BTN_TC_ID == TC1_ID
#define BTN_TC_CC_REG   TC1_CCR
#define BTN_TC_ID_REG   TC1_IDR
#define BTN_TC_S_REG    TC1_SR
#define BTN_TC_CM_REG   TC1_CMR
#define BTN_TC_IE_REG   TC1_IER
#define BTN_TC_RC_REG   TC1_RC
#define sig_BTN_TC      sig_TC1
#elif BTN_TC_ID == TC2_ID
#define BTN_TC_CC_REG   TC2_CCR
#define BTN_TC_ID_REG   TC2_IDR
#define BTN_TC_S_REG    TC2_SR
#define BTN_TC_CM_REG   TC2_CMR
#define BTN_TC_IE_REG   TC2_IER
#define BTN_TC_RC_REG   TC2_RC
#define sig_BTN_TC      sig_TC2
#endif

/*!
 * \brief Button event queue.
 */
static HANDLE btn_que;

static volatile u_int btn_pressed;
static volatile u_int btn_code;
static volatile u_int btn_prev;
static volatile u_int btn_repeat;

/*!
 * \brief Button scan timer interrupt handler.
 */
static void ScanTimerInterrupt(void *arg)
{
    /* Read the negated GPIO line status. Pressing a button drives the pin low. */
    btn_code = ~inr(BTN_PIO_PDS_REG) & (BTN_DOWN | BTN_MENU | BTN_UP);
    if (btn_code) {
        /* A button has been or still is pressed. */
        if (btn_code != btn_prev) {
            /* This is a new button. Immediately post an event. */
            btn_pressed = btn_code;
            NutEventPostFromIrq(&btn_que);
            /* Initialize the repeat rate. */
            btn_repeat = BTN_REPEAT_FIRST;
        }
        else if (btn_repeat == 0) {
            /* Button is still pressed and the repeat time elapsed. 
               Post a new event. */
            NutEventPostFromIrq(&btn_que);
            /* Re-initialize the repeat rate. */
            btn_repeat = BTN_REPEAT_NEXT;
        }
        else {
            /* Button is still pressed. Wait until repeat time elapses. */
            btn_repeat--;
        }
    }
    /* Keep the last code, so we are able to distinguish whether the user
       is still pressing the same or selected a new button. */
    btn_prev = btn_code;
}

/*!
 * \brief Initialize the button interface.
 *
 * Configures and starts a continously running timer interrupt, which scans
 * three GPIO lines. A low level on any line indicates, that the related
 * button has been pressed by the user.
 *
 * The application must call ButtonRead() to read the code of a pressed button.
 */
void ButtonInit(void)
{
    int dummy;

    /* Enable scan timer clock. */
    outr(PMC_PCER, _BV(BTN_TC_ID) | _BV(BTN_PIO_ID));
    /* Disable the Clock Counter */
    outr(BTN_TC_CC_REG, TC_CLKDIS);
    /* Disable all interrupts */
    outr(BTN_TC_ID_REG, 0xFFFFFFFF);
    /* Clear the status register. */
    dummy = inr(BTN_TC_S_REG);
    /* Select divider and compare trigger */
    outr(BTN_TC_CM_REG, TC_CLKS_MCK32 | TC_CPCTRG);
    /* Enable the Clock counter */
    outr(BTN_TC_CC_REG, TC_CLKEN);
    /* Validate the RC compare interrupt */
    outr(BTN_TC_IE_REG, TC_CPCS);

    /* Register timer interrupt handler. */
    NutRegisterIrqHandler(&sig_BTN_TC, ScanTimerInterrupt, 0);
    /* Set to lowest priority. */
    NutIrqSetPriority(&sig_BTN_TC, 0);

    /* Enable timer interrupts */
    NutIrqEnable(&sig_BTN_TC);

    /* Set compare value for specified scan frequency. */
#if defined(AT91_PLL_MAINCK)
    outr(BTN_TC_RC_REG, At91GetMasterClock() / (32 * BTN_SCAN_FREQ));
#else
    outr(BTN_TC_RC_REG, NutGetCpuClock() / (32 * BTN_SCAN_FREQ));
#endif

    /* Initialize GPIO lines for buttons. */
    outr(BTN_PIO_PE_REG, BTN_MENU | BTN_UP | BTN_DOWN);
    outr(BTN_PIO_OD_REG, BTN_MENU | BTN_UP | BTN_DOWN);
    outr(BTN_PIO_PUE_REG, BTN_MENU | BTN_UP | BTN_DOWN);

    /* Software trigger starts the scan timer. */
    outr(BTN_TC_CC_REG, TC_SWTRG);
}

/*!
 * \brief Wait until the user pressed a button.
 *
 * \param tmo Maximum number of milliseconds to wait.
 *
 * \return ASCII code of the button, either 'M', 'D' or 'U'. If no button
 *         had been pressed within the given time, then zero is returned.
 */
char ButtonRead(u_long tmo)
{
    if (NutEventWait(&btn_que, tmo) == 0) {
        u_int pressed = btn_pressed;

        if (pressed & BTN_MENU) {
            return 'M';
        }
        if (pressed & BTN_DOWN) {
            return 'D';
        }
        if (pressed & BTN_UP) {
            return 'U';
        }
    }
    return 0;
}
