/*
 * Copyright (C) 2003-2007 by egnite Software GmbH. All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. Neither the name of the copyright holders nor the names of
 *    contributors may be used to endorse or promote products derived
 *    from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY EGNITE SOFTWARE GMBH AND CONTRIBUTORS
 * ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 * FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL EGNITE
 * SOFTWARE GMBH OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS
 * OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED
 * AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF
 * THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * For additional information see http://www.ethernut.de/
 *
 */

/*!
 * \file config.c
 * \brief User interface.
 *
 * \verbatim
 *
 * $Log$
 *
 * \endverbatim
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <dev/term.h>

#include <sys/thread.h>
#include <sys/timer.h>
#include <sys/event.h>

#include <arpa/inet.h>

#include "config.h"
#include "station.h"
#include "buttons.h"
#include "userif.h"

#ifndef DISPLAY_LINES
#define DISPLAY_LINES       2
#endif

#ifndef DISPLAY_COLUMNS
#define DISPLAY_COLUMNS     16
#endif

#ifndef DISPLAY_VCOLUMNS
#ifdef AT91SAM7X_EK
#define DISPLAY_VCOLUMNS    40
#else
#define DISPLAY_VCOLUMNS    80
#endif
#endif

typedef struct {
    char dln_status[DISPLAY_VCOLUMNS];
    char dln_msg[DISPLAY_VCOLUMNS];
    u_int dln_step;
    u_int dln_len;
    u_short dln_msgticks;
} DISPLAYLINE;

typedef struct {
    FILE *dpi_stream;
    u_char dpi_scrolling;   /*!< . */
    char *dpi_sptr;
    DISPLAYLINE dpi_line[DISPLAY_LINES];
} DISPLAYINFO;

DISPLAYINFO display;

static HANDLE updevt;

static int d_timevalid; 
static STATIONINFO *d_sip;
static RECEIVERINFO *d_rip;
static PLAYERINFO *d_pip;

/*!
 * \brief Refresh the LCD.
 */
static void DisplayRefresh(void)
{
    char *line;
    u_char ln;

    for (ln = 0; ln < DISPLAY_LINES; ln++) {
        /*
         * Determine if we display a message or just the current status.
         */
        if (display.dpi_line[ln].dln_msgticks) {
            /* Displaying a message. */
            line = display.dpi_line[ln].dln_msg;
            display.dpi_line[ln].dln_msgticks--;
        } else {
            /* Displaying the status. */
            line = display.dpi_line[ln].dln_status;
        }

        /* 
         * Re-initialize scrolling, if the line length changed.
         */
        if (display.dpi_line[ln].dln_len != strlen(line)) {
            display.dpi_sptr = NULL;
            display.dpi_scrolling = DISPLAY_LINES;
            display.dpi_line[ln].dln_len = strlen(line);
            display.dpi_line[ln].dln_step = strlen(line);
        }

        /* Position the cursor. */
        fprintf(display.dpi_stream, ESC_POS "%c" "\x20", ln + 32);

        if (display.dpi_line[ln].dln_len <= DISPLAY_COLUMNS) {
            /* No scrolling required. */
            fputs(line, display.dpi_stream);
            if (display.dpi_line[ln].dln_len < DISPLAY_COLUMNS) {
                fputs(ESC_CLREOL, display.dpi_stream);
            }
        } else {
            /* Line is larger than our physical display. Scroll horizontally. */
            if (display.dpi_line[ln].dln_step == display.dpi_line[ln].dln_len) {
                if (display.dpi_scrolling == DISPLAY_LINES) {
                    /* Scroll this line, if no line is currently scrolling. */
                    display.dpi_scrolling = ln;
                }
                if (display.dpi_scrolling == ln) {
                    /* This line is scrolling. */
                    if (display.dpi_sptr == NULL) {
                        display.dpi_sptr = line;
                        fprintf(display.dpi_stream, "%.16s", line);
                    }
                    else {
                        display.dpi_sptr++;
                        if (strlen(display.dpi_sptr) <= DISPLAY_COLUMNS) {
                            display.dpi_line[ln].dln_step++;
                        }
                        fprintf(display.dpi_stream, "%.16s", display.dpi_sptr);
                    }
                }
            } else if (display.dpi_line[ln].dln_step++ == 0) {
                fprintf(display.dpi_stream, "%.16s", line);
                display.dpi_sptr = NULL;
                display.dpi_scrolling = DISPLAY_LINES;
            } else if (display.dpi_line[ln].dln_step >= display.dpi_line[ln].dln_len + DISPLAY_COLUMNS) {
                display.dpi_line[ln].dln_step = 0;
            }
        }
    }
}

void UserIfMainMenu(void)
{
    char key;
    u_char want = radio.rc_rstation;
    u_char got = want + 1;

    for (;;) {
        if (want != got) {
            if (want >= MAXNUM_STATIONS) {
                want = 0;
            }
            got = want;
            fprintf(display.dpi_stream, "\n%s", inet_ntoa(station[got].rs_ip));
            fprintf(display.dpi_stream, "\nPrev Select Next");
        }
        key = ButtonRead(5000);
        if (key == 'D') {
            do {
                if (want == 0) {
                    want = MAXNUM_STATIONS;
                }
                want--;
            } while (station[want].rs_port == 0);
        }
        else if (key == 'U') {
            do {
                want++;
                if (want >= MAXNUM_STATIONS) {
                    want = 0;
                }
            } while (station[want].rs_port == 0);
        }
        else if (key == 'M') {
            break;
        }
        else {
            want = radio.rc_rstation;
        }
    }
    if (want != radio.rc_rstation) {
        DisplayMessage(1, 5, "Please wait");
        radio.rc_rstation = want;
    }
}

/*
 * Background thread for user interface.
 */
THREAD(UserIfThread, arg)
{
    char key;

    fputs(ESC_CURSOROFF, display.dpi_stream);
    NutThreadSetPriority(128);
    
    for (;;) {
        key = ButtonRead(250);
        if (key == 'D') {
            radio.rc_rvolume--;
        }
        else if (key == 'M') {
            UserIfMainMenu();
        }
        else if (key == 'U') {
            radio.rc_rvolume++;
        }
        else {
            DisplayRefresh();
        }
    }
}

void UserIfStation(STATIONCONF * rsp)
{
    if (rsp->rs_port) {
        if (rsp->rs_symbol && rsp->rs_symbol[0])
            strncpy(display.dpi_line[0].dln_status, rsp->rs_symbol, DISPLAY_VCOLUMNS);
        else
            strncpy(display.dpi_line[0].dln_status, inet_ntoa(rsp->rs_ip), DISPLAY_VCOLUMNS);
    } else
        display.dpi_line[0].dln_status[0] = 0;
}

/*!
 * \brief Display a specified status.
 *
 * \param status Status to be displayed.
 */
void DisplayStatus(u_char status)
{
    if (radio.rc_cstatus != status) {
        if (status == DIST_FORCE)
            status = radio.rc_cstatus;
        else
            radio.rc_cstatus = status;

        if (status == DIST_NONE) {
            strcpy(display.dpi_line[0].dln_status, "Internet Radio");
            strcpy(display.dpi_line[1].dln_status, "Version ");
            strcat(display.dpi_line[1].dln_status, VERSION);
            NutEventPost(&updevt);
        } else if (status == DIST_DEAD) {
            sprintf(display.dpi_line[0].dln_status, "Station %03u", radio.rc_cstation);
            strcpy(display.dpi_line[1].dln_status, "not available");
            NutEventPost(&updevt);
        } else if (status == DIST_CONNECTING) {
            UserIfStation(&station[radio.rc_cstation]);
            strcpy(display.dpi_line[1].dln_status, "Connecting...");
            NutEventPost(&updevt);
        } else if (status == DIST_CONNECTED) {
            UserIfStation(&station[radio.rc_cstation]);
            if (d_rip) {
                SHOUTCASTINFO *sci = (SHOUTCASTINFO *) d_rip->ri_bcast;
                if (sci) {
                    if (sci->sci_metatitle && sci->sci_metatitle[0]) {
                        strncpy(display.dpi_line[1].dln_status, sci->sci_metatitle, DISPLAY_VCOLUMNS);
                    }
                    else if (d_sip && d_sip->si_genre && d_sip->si_genre[0]) {
                        strncpy(display.dpi_line[1].dln_status, d_sip->si_genre, DISPLAY_VCOLUMNS);
                    }
                    NutEventPost(&updevt);
                }
            }
        }
    }
}

/*!
 * \brief Display a specified text.
 *
 * \param row  Row position of the message.
 * \param secs Number of seconds to display the message. Set to 0 for
 *             permanent display.
 * \param fmt  Format string containing conversion specifications.
 */
void DisplayMessage(u_char row, u_char secs, CONST char *fmt, ...)
{
    va_list ap;

    va_start(ap, fmt);
    if (secs) {
        vsprintf(display.dpi_line[row].dln_msg, fmt, ap);
        display.dpi_line[row].dln_msgticks = secs * 4;
    } else
        vsprintf(display.dpi_line[row].dln_status, fmt, ap);
    va_end(ap);

    NutEventPost(&updevt);
}

/*!
 * \brief Start background thread for display updates.
 *
 * \param name Display device name.
 *
 * \return 0 on success or -1 in case of a failure.
 */
int UserIfInit(char *name)
{
    /* Initialize button interface. */
    ButtonInit();

    if ((display.dpi_stream = fopen(name, "w")) == 0) {
        return -1;
    }
    display.dpi_scrolling = DISPLAY_LINES;

    if (NutThreadCreate("displ", UserIfThread, 0, 512) == 0) {
        return -1;
    }
    radio.rc_cstatus = DIST_NONE;
    DisplayStatus(DIST_FORCE);

    return 0;
}

int UserIfSetStationInfo(STATIONINFO *sip)
{
    d_sip = sip;
    return 0;
}

int UserIfSetReceiverInfo(RECEIVERINFO *rip)
{
    d_rip = rip;
    return 0;
}

int UserIfSetPlayerInfo(PLAYERINFO *pip)
{
    d_pip = pip;
    return 0;
}
